/*
   This file is part of GNOME Klotski.

   GNOME Klotski is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   GNOME Klotski is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with GNOME Klotski.  If not, see <https://www.gnu.org/licenses/>.
*/

button.unfullscreen-button {
  margin:6px;
}

popover#puzzles-popover {
  padding: 10px;
}

stack#stack-packs > label {
  font-weight: bold;
}

window:not(.hc-theme) .treeview-container treeview {
  background-color: transparent;
}
window:not(.hc-theme) .treeview-container treeview:backdrop {
  color: @theme_unfocused_fg_color;
}
window:not(.hc-theme) .treeview-container treeview:hover {
  background-color: @theme_bg_color;
}
window:not(.hc-theme) .treeview-container treeview:hover:backdrop {
  background-color: @theme_unfocused_bg_color;
}
window:not(.hc-theme) .treeview-container treeview:selected,
window:not(.hc-theme) .treeview-container treeview:selected:hover {
  background-color: @theme_selected_bg_color;
  color: @theme_selected_fg_color;
}
window:not(.hc-theme) .treeview-container treeview:selected:backdrop,
window:not(.hc-theme) .treeview-container treeview:selected:backdrop:hover {
  background-color: @theme_unfocused_selected_bg_color;
  color: @theme_unfocused_selected_fg_color;
}

treeview.toggle-like-pixbuf {
  margin-right:22px;      /* TODO hack 1/2 */
}
treeview.toggle-like-pixbuf check {
  min-width:22px;         /* TODO hack 2/2 */
  background:transparent;
  border:none;
  -gtk-icon-source: none; /* TODO report bug about HighContrast, that shouldn’t set this */
}
treeview.toggle-like-pixbuf check:checked {
  -gtk-icon-source: -gtk-icontheme("gtk-yes"); /* TODO maybe deprecated; replace with CSS? */
}

/*\
* * the end
\*/
